from functools import partial
from typing import TypeVar

import colander
from pyramid.view import view_config, view_defaults
from sqlalchemy.orm import subqueryload

from c2cgeoform.ext.deform_ext import RelationCheckBoxListWidget
from c2cgeoform.schema import GeoFormManyToManySchemaNode, GeoFormSchemaNode, manytomany_validator
from c2cgeoform.views.abstract_views import AbstractViews, ItemAction, ListField, UserMessage

from ..i18n import _
from ..models.c2cgeoform_demo import Excavation, Situation

_list_field = partial(ListField, Excavation)

base_schema = GeoFormSchemaNode(Excavation)
base_schema.add_before(
    "contact_persons",
    colander.SequenceSchema(
        GeoFormManyToManySchemaNode(Situation, None),
        name="situations",
        title="Situations",
        widget=RelationCheckBoxListWidget(Situation, "id", "name", order_by="name"),
        validator=manytomany_validator,
    ),
)
base_schema.add_unique_validator(Excavation.reference_number, Excavation.hash)


T = TypeVar("T")


@view_defaults(match_param="table=excavations")
class ExcavationViews(AbstractViews[T]):
    _model = Excavation
    _base_schema = base_schema
    _id_field = "hash"
    _geometry_field = "work_footprint"

    _list_fields = [
        _list_field("reference_number"),
        _list_field("request_date"),
        _list_field("description"),
        _list_field("location_town"),
        _list_field("responsible_company"),
        _list_field(
            "situations",
            renderer=lambda excavation: ", ".join([s.name for s in excavation.situations]),
            filter_column=Situation.name,
        ),
    ]

    MSG_COL = {**AbstractViews.MSG_COL, "error": UserMessage(_("This is an error"), "alert-danger")}

    def _base_query(self):
        return (
            super()
            ._base_query()
            .distinct()
            .outerjoin(Excavation.situations)
            .options(subqueryload(Excavation.situations))
        )

    @view_config(route_name="c2cgeoform_index", renderer="../templates/index.jinja2")
    def index(self):
        return super().index()

    @view_config(route_name="c2cgeoform_grid", renderer="json")
    def grid(self):
        return super().grid()

    def _grid_actions(self):
        return super()._grid_actions() + [
            ItemAction(
                name="action_map",
                label=_("Map"),
                css_class="btn btn-primary btn-map",
                url=self._request.route_url("c2cgeoform_map"),
            )
        ]

    def _item_actions(self, item, readonly=False):
        actions = super()._item_actions(item, readonly)
        if item is not None:
            actions.append(
                ItemAction(
                    name="error",
                    label=_("Show an error"),
                    icon="glyphicon glyphicon-alert",
                    url=self._request.route_url(
                        self._request.matched_route.name,
                        **self._request.matchdict,
                        _query=[("msg_col", "error")],
                    ),
                )
            )
        return actions

    @view_config(route_name="c2cgeoform_map", renderer="../templates/map.jinja2")
    def map(self):
        return super().map({})

    @view_config(route_name="c2cgeoform_geojson", renderer="json")
    def geojson(self):
        return super().geojson()

    @view_config(route_name="c2cgeoform_item", request_method="GET", renderer="../templates/edit.jinja2")
    def edit(self):
        return super().edit()

    @view_config(
        route_name="c2cgeoform_item_duplicate", request_method="GET", renderer="../templates/edit.jinja2"
    )
    def duplicate(self):
        return super().duplicate()

    @view_config(route_name="c2cgeoform_item", request_method="DELETE", renderer="json")
    def delete(self):
        return super().delete()

    @view_config(route_name="c2cgeoform_item", request_method="POST", renderer="../templates/edit.jinja2")
    def save(self):
        return super().save()



from pkg_resources import resource_filename
from pyramid.config import Configurator

import c2cgeoform
from c2cgeoform.settings import apply_local_settings

search_paths = (
    resource_filename("{{cookiecutter.package}}", "templates/widgets"),
) + c2cgeoform.default_search_paths
c2cgeoform.default_search_paths = search_paths


def main(global_config, **settings):
    """This function returns a Pyramid WSGI application."""
    apply_local_settings(settings)

    config = Configurator(
        settings=settings, locale_negotiator="{{cookiecutter.package}}.i18n.locale_negotiator"
    )
    config.include("pyramid_chameleon")
    config.include("pyramid_mako")
    config.include("pyramid_jinja2")

    config.include("c2cgeoform")

    config.include(".models")
    config.include(".routes")

    config.add_translation_dirs("locale")

    config.scan()

    return config.make_wsgi_app()



from colander import null

from c2cgeoform.models import DBSession
from c2cgeoform.tests import DatabaseTestCase

from .models_test import EmploymentStatus, Person, Tag


class TestRelationSelectWidget(DatabaseTestCase):
    def test_serialize(self):
        from c2cgeoform.ext.deform_ext import RelationSelectWidget

        widget = RelationSelectWidget(EmploymentStatus, "id", "name")

        renderer = DummyRenderer()
        field = DummyField(None, renderer=renderer)
        widget.populate(DBSession, None)
        widget.serialize(field, null)
        self.assertEqual(renderer.kw["values"], _convert_values(widget.values))

        first_value = renderer.kw["values"][0]
        self.assertEqual("0", first_value[0])
        self.assertEqual("Worker", first_value[1])

    def test_serialize_default_value(self):
        from c2cgeoform.ext.deform_ext import RelationSelectWidget

        widget = RelationSelectWidget(EmploymentStatus, "id", "name", ("", "- Select -"))

        renderer = DummyRenderer()
        field = DummyField(None, renderer=renderer)
        widget.populate(DBSession, None)
        widget.serialize(field, null)
        self.assertEqual(renderer.kw["values"], _convert_values(widget.values))

        # first the default value
        first_value = renderer.kw["values"][0]
        self.assertEqual("", first_value[0])
        self.assertEqual("- Select -", first_value[1])

        # then the values loaded from the db
        snd_value = renderer.kw["values"][1]
        self.assertEqual("0", snd_value[0])
        self.assertEqual("Worker", snd_value[1])

    def test_serialize_order_by(self):
        from c2cgeoform.ext.deform_ext import RelationSelectWidget

        widget = RelationSelectWidget(EmploymentStatus, "id", "name", order_by="name")

        renderer = DummyRenderer()
        field = DummyField(None, renderer=renderer)
        widget.populate(DBSession, None)
        widget.serialize(field, null)
        self.assertEqual(renderer.kw["values"], _convert_values(widget.values))

        first_value = renderer.kw["values"][0]
        self.assertEqual("3", first_value[0])
        self.assertEqual("Director", first_value[1])


class TestRelationSelect2Widget(DatabaseTestCase):
    def test_serialize_default_value(self):
        from c2cgeoform.ext.deform_ext import RelationSelect2Widget

        widget = RelationSelect2Widget(EmploymentStatus, "id", "name", ("", "- Select -"))

        renderer = DummyRenderer()
        field = DummyField(None, renderer=renderer)
        widget.populate(DBSession, None)
        widget.serialize(field, null)
        self.assertEqual(renderer.kw["values"], _convert_values(widget.values))

        # first the default value
        first_value = renderer.kw["values"][0]
        self.assertEqual("", first_value[0])
        self.assertEqual("- Select -", first_value[1])

        # then the values loaded from the db
        snd_value = renderer.kw["values"][1]
        self.assertEqual("0", snd_value[0])
        self.assertEqual("Worker", snd_value[1])


class TestRelationRadioChoiceWidget(DatabaseTestCase):
    def test_serialize_default_value(self):
        from c2cgeoform.ext.deform_ext import RelationRadioChoiceWidget

        widget = RelationRadioChoiceWidget(EmploymentStatus, "id", "name")

        renderer = DummyRenderer()
        field = DummyField(None, renderer=renderer)
        widget.populate(DBSession, None)
        widget.serialize(field, null)
        self.assertEqual(renderer.kw["values"], _convert_values(widget.values))

        # first the default value
        first_value = renderer.kw["values"][0]
        self.assertEqual("0", first_value[0])
        self.assertEqual("Worker", first_value[1])


class TestRelationCheckBoxListWidget(DatabaseTestCase):
    def test_serialize_empty(self):
        from c2cgeoform.ext.deform_ext import RelationCheckBoxListWidget

        widget = RelationCheckBoxListWidget(Tag, "id", "name")
        renderer = DummyRenderer()

        field = _get_field("tags", renderer)
        widget.populate(DBSession, None)
        widget.serialize(field, null)
        self.assertEqual(renderer.kw["values"], _convert_values(widget.values))
        self.assertEqual(renderer.kw["cstruct"], [])

        first_value = renderer.kw["values"][0]
        self.assertEqual("0", first_value[0])
        self.assertEqual("Tag A", first_value[1])

    def test_serialize(self):
        from c2cgeoform.ext.deform_ext import RelationCheckBoxListWidget

        widget = RelationCheckBoxListWidget(Tag, "id", "name")
        renderer = DummyRenderer()

        field = _get_field("tags", renderer)
        widget.populate(DBSession, None)
        objs = [{"id": "0"}, {"id": "2"}]

        widget.serialize(field, objs)
        self.assertEqual(renderer.kw["values"], _convert_values(widget.values))
        self.assertEqual(renderer.kw["cstruct"], ["0", "2"])

        first_value = renderer.kw["values"][0]
        self.assertEqual("0", first_value[0])
        self.assertEqual("Tag A", first_value[1])

    def test_serialize_wrong_mapping(self):
        from c2cgeoform.ext.deform_ext import RelationCheckBoxListWidget

        widget = RelationCheckBoxListWidget(EmploymentStatus, "id", "name")
        renderer = DummyRenderer()

        field = _get_field("tags", renderer)
        widget.populate(DBSession, None)
        objs = [{"bad_column": "101"}, {"bad_column": "102"}]

        with self.assertRaises(KeyError):
            widget.serialize(field, objs)

    def test_deserialize_empty(self):
        from c2cgeoform.ext.deform_ext import RelationCheckBoxListWidget

        widget = RelationCheckBoxListWidget(Tag, "id", "name")
        renderer = DummyRenderer()

        field = _get_field("tags", renderer)
        widget.populate(DBSession, None)
        result = widget.deserialize(field, null)
        self.assertEqual(result, [])

    def test_deserialize(self):
        from c2cgeoform.ext.deform_ext import RelationCheckBoxListWidget

        widget = RelationCheckBoxListWidget(Tag, "id", "name")
        renderer = DummyRenderer()

        field = _get_field("tags", renderer)
        widget.populate(DBSession, None)
        result = widget.deserialize(field, ["1", "2"])
        self.assertEqual(result, [{"id": "1"}, {"id": "2"}])


class TestRelationMultiSelect2Widget(DatabaseTestCase):
    def test_serialize_empty(self):
        from c2cgeoform.ext.deform_ext import RelationSelect2Widget

        widget = RelationSelect2Widget(Tag, "id", "name", multiple=True)
        renderer = DummyRenderer()

        field = _get_field("tags", renderer)
        widget.populate(DBSession, None)
        widget.serialize(field, null)
        self.assertEqual(renderer.kw["values"], _convert_values(widget.values))
        self.assertEqual(renderer.kw["cstruct"], [])

        first_value = renderer.kw["values"][0]
        self.assertEqual("0", first_value[0])
        self.assertEqual("Tag A", first_value[1])

    def test_serialize(self):
        from c2cgeoform.ext.deform_ext import RelationSelect2Widget

        widget = RelationSelect2Widget(Tag, "id", "name", multiple=True)
        renderer = DummyRenderer()

        field = _get_field("tags", renderer)
        widget.populate(DBSession, None)
        objs = [{"id": "0"}, {"id": "2"}]

        widget.serialize(field, objs)
        self.assertEqual(renderer.kw["values"], _convert_values(widget.values))
        self.assertEqual(renderer.kw["cstruct"], ["0", "2"])

        first_value = renderer.kw["values"][0]
        self.assertEqual("0", first_value[0])
        self.assertEqual("Tag A", first_value[1])

    def test_serialize_wrong_mapping(self):
        from c2cgeoform.ext.deform_ext import RelationSelect2Widget

        widget = RelationSelect2Widget(EmploymentStatus, "id", "name", multiple=True)
        renderer = DummyRenderer()

        field = _get_field("tags", renderer)
        widget.populate(DBSession, None)
        objs = [{"bad_column": "101"}, {"bad_column": "102"}]

        with self.assertRaises(KeyError):
            widget.serialize(field, objs)

    def test_deserialize_empty(self):
        from c2cgeoform.ext.deform_ext import RelationSelect2Widget

        widget = RelationSelect2Widget(Tag, "id", "name", multiple=True)
        renderer = DummyRenderer()

        field = _get_field("tags", renderer)
        widget.populate(DBSession, None)
        result = widget.deserialize(field, null)
        self.assertEqual(result, [])

    def test_deserialize(self):
        from c2cgeoform.ext.deform_ext import RelationSelect2Widget

        widget = RelationSelect2Widget(Tag, "id", "name", multiple=True)
        renderer = DummyRenderer()

        field = _get_field("tags", renderer)
        widget.populate(DBSession, None)
        result = widget.deserialize(field, ["1", "2"])
        self.assertEqual(result, [{"id": "1"}, {"id": "2"}])


def _convert_values(values_tuple):
    return [(str(key), label) for (key, label) in values_tuple]


def _get_field(name, renderer):
    from colanderalchemy import SQLAlchemySchemaNode
    from deform import Form

    form = Form(SQLAlchemySchemaNode(Person), renderer=renderer)

    for field in form.children:
        if field.name == name:
            return field

    raise RuntimeError("Field not found")


class DummyRenderer:
    """A dummy renderer, borrowed from the deform tests."""

    def __init__(self, result=""):
        self.result = result

    def __call__(self, template, **kw):
        self.template = template
        self.kw = kw
        return self.result


class DummyField:
    """A dummy field, borrowed from the deform tests."""

    default = None
    error = None
    children = ()
    title = "title"
    description = "description"
    name = "name"
    cloned = False
    oid = "deformField1"
    required = True
    cstruct = null

    def __init__(self, schema=None, renderer=None, translations=None):
        self.schema = schema
        self.renderer = renderer
        self.translations = translations

    def clone(self):
        self.cloned = True
        return self

    def deserialize(self, pstruct):
        return self.widget.deserialize(self, pstruct)

    def translate(self, term):
        if self.translations is None:
            return term
        return self.translations.get(term, term)

    def render_template(self, template, **kw):
        return self.renderer(template, **kw)



import colander
import deform
from sqlalchemy import Boolean, Column, ForeignKey, Integer, Table, Text
from sqlalchemy.orm import relationship

from c2cgeoform.ext.deform_ext import RelationSelect2Widget
from c2cgeoform.models import Base


class EmploymentStatus(Base):
    __tablename__ = "tests_employment_status"

    id = Column(Integer, primary_key=True)
    name = Column(Text, nullable=False)


class Phone(Base):
    __tablename__ = "tests_phones"

    id = Column(Integer, primary_key=True, info={"colanderalchemy": {"widget": deform.widget.HiddenWidget()}})
    number = Column(Text, nullable=False, info={"colanderalchemy": {"title": "Phone number"}})
    person_id = Column(
        Integer,
        ForeignKey("tests_persons.id"),
        info={"colanderalchemy": {"widget": deform.widget.HiddenWidget()}},
    )
    verified = Column(Boolean)


class Tag(Base):
    __tablename__ = "tests_tags"

    id = Column(Integer, primary_key=True)
    name = Column(Text, nullable=False)


person_tag = Table(
    "person_tag",
    Base.metadata,
    Column("tag_id", Integer, ForeignKey("tests_tags.id"), primary_key=True),
    Column("person_id", Integer, ForeignKey("tests_persons.id"), primary_key=True),
)


class Person(Base):
    __tablename__ = "tests_persons"
    __colanderalchemy_config__ = {"title": "A Person", "description": "Tell us about you."}
    __c2cgeoform_config__ = {"duplicate": True}

    id = Column(
        Integer,
        primary_key=True,
        info={"colanderalchemy": {"title": "ID", "widget": deform.widget.HiddenWidget()}},
    )
    hash = Column(Text, unique=True)
    name = Column(
        Text,
        nullable=False,
        info={
            "colanderalchemy": {
                "title": "Your name",
            },
            "c2cgeoform": {"duplicate": True},
        },
    )
    first_name = Column(
        Text,
        nullable=False,
        info={"colanderalchemy": {"title": "Your first name"}, "c2cgeoform": {"duplicate": True}},
    )
    age = Column(
        Integer,
        info={
            "colanderalchemy": {
                "title": "Your age",
                "validator": colander.Range(
                    18,
                ),
            }
        },
    )
    phones = relationship(
        Phone,
        cascade="all, delete-orphan",
        info={
            "colanderalchemy": {
                "title": "Phone numbers",
            }
        },
    )
    tags = relationship(
        "Tag",
        secondary=person_tag,
        cascade="save-update,merge,refresh-expire",
        info={
            "colanderalchemy": {
                "title": "Tags",
                "widget": RelationSelect2Widget(Tag, "id", "name", order_by="name", multiple=True),
                "includes": ["id"],
            }
        },
    )

    validated = Column(Boolean, info={"colanderalchemy": {"title": "Validation", "label": "Validated"}})



import re

import pytest
from pyramid.testing import DummyRequest
from {{cookiecutter.package}}.models.c2cgeoform_demo import Excavation, Situation
from {{cookiecutter.package}}.views.excavation import ExcavationViews


@pytest.mark.usefixtures("transact", "test_app")
class TestExcavationViews():
    def check_checkboxes_list_widget(self, resp, all_situations, excavation):
        form_group = resp.html.select_one('.item-situations')
        situations_labels = form_group.select('.checkbox label')
        for i, situation in enumerate(all_situations):
            label = situations_labels[i]
            input = label.select_one('input')
            assert [situation.name] == list(label.stripped_strings)
            assert str(situation.id) == input['value']
            if situation in excavation.situations:
                assert 'checked' == input['checked']
                assert resp.form.get('situations', index=i).checked
            else:
                assert 'checked' not in input
                assert not resp.form.get('situations', index=i).checked

    def _excavation_by_desc(self, dbsession, desc):
        return dbsession.query(Excavation). \
            filter(Excavation.description == desc). \
            one_or_none()

    def _excavation_by_ref(self, dbsession, ref):
        return dbsession.query(Excavation). \
            filter(Excavation.reference_number == ref). \
            one_or_none()

    @pytest.mark.skip(reason="Not working with Pyramid 2.0")
    def test_index(self, test_app):
        resp = test_app.get("/excavations", status=200)

        resp_tmp = resp.click(verbose=True, href="language=en")
        resp_en = resp_tmp.follow()
        html_en = resp_en.html
        news_en = html_en.select('a[href$="/excavations/new"]')
        assert 1 == len(news_en)
        assert ["New"] == list(news_en[0].stripped_strings)

        reference_numbers = html_en.select("th[data-field=reference_number]")
        assert 1 == len(reference_numbers)
        assert "Reference Number" == reference_numbers[0].string

        resp_tmp = resp_en.click(verbose=True, href="language=fr")
        resp_fr = resp_tmp.follow()
        html_fr = resp_fr.html
        news_fr = html_fr.select('a[href$="/excavations/new"]')
        assert 1 == len(news_fr)
        assert ["Nouveau"] == list(news_fr[0].stripped_strings)

    def test_grid(self, test_app, dbsession):
        json = test_app.post(
            "/excavations/grid.json",
            params={
                "offset": 0,
                "limit": 10,
                "sort": "reference_number",
                "order": "asc"
            },
            status=200
        ).json

        assert 10 == len(json["rows"])
        assert 100 == json["total"]

        row = json["rows"][5]
        exc = self._excavation_by_ref(dbsession, row["reference_number"])
        assert exc.hash == row["_id_"]
        assert exc.description == row["description"]
        assert exc.location_town == row["location_town"]
        assert exc.request_date.strftime("%Y-%m-%d") == row["request_date"]
        assert exc.responsible_company == row["responsible_company"]
        assert ", ".join([situation.name for situation in exc.situations]) == row["situations"]

    def test_grid_filter(self, test_app):
        json = test_app.post(
            "/excavations/grid.json",
            params={
                "offset": 0,
                "limit": 10,
                "sort": "reference_number",
                "order": "asc",
                "search": "Road"
            },
            status=200
        ).json

        assert 10 == len(json["rows"])
        assert 49 == json["total"]

        row = json["rows"][0]
        assert "Road" in row["situations"]

    def test_edit_then_save(self, dbsession, test_app):
        exc = self._excavation_by_ref(dbsession, "ref0001")
        resp = test_app.get("/excavations/{}".format(exc.hash), status=200)

        form = resp.form
        assert exc.reference_number == form['reference_number'].value
        assert exc.hash == form['hash'].value
        assert str(exc.id) == form.fields.get('id')[0].value
        assert str(exc.request_date) == form['date'].value
        assert exc.description == form["description"].value
        assert exc.motif == form["motif"].value
        assert exc.contact_persons[0].first_name == form["first_name"].value
        assert exc.contact_persons[0].last_name == form["last_name"].value
        assert str(exc.contact_persons[0].id) == form.fields.get('id')[1].value
        assert str(exc.contact_persons[0].permission_id) == form['permission_id'].value
        assert exc.contact_persons[0].permission_id == exc.id
        assert str(exc.location_district_id) == form["location_district_id"].value
        assert exc.location_street == form["location_street"].value
        assert exc.location_postal_code == form["location_postal_code"].value
        assert exc.location_town == form["location_town"].value
        assert str(exc.address_id or "") == form["address_id"].value
        assert str(exc.location_position or "") == form["location_position"].value
        assert exc.responsible_title == form["responsible_title"].value
        assert exc.responsible_name == form["responsible_name"].value
        assert exc.responsible_first_name == form["responsible_first_name"].value
        assert exc.responsible_mobile == form["responsible_mobile"].value
        assert exc.responsible_mail == form["responsible_mail"].value
        assert exc.responsible_company == form["responsible_company"].value
        assert "true" if exc.validated else "false" == form["validated"].value
        assert str(exc.work_footprint or "") == form["work_footprint"].value
        # assert form["photos"].value == exc.photos

        all_situations = dbsession.query(Situation).order_by(Situation.name).all()
        self.check_checkboxes_list_widget(resp, all_situations, exc)

        form["description"] = "New description"
        form["motif"] = "New motif"
        form["location_district_id"] = 4  # name="La Claie-aux-Moines"
        form["location_street"] = "rue des écoles"
        form["location_postal_code"] = '73000'
        form["location_town"] = "Beçanson"
        form["address_id"] = 4  # label="Lugano"
        form["location_position"] = ''
        form["responsible_title"] = "mr"
        form["responsible_name"] = "lapin"
        form["responsible_first_name"] = "malin"
        form["responsible_mobile"] = "06.75.65.83"
        form["responsible_mail"] = "lapin.malin@lugano.ch"
        form["responsible_company"] = "csc"
        form["validated"] = False
        form["work_footprint"] = ''
        form["first_name"] = 'Raoul'
        form["last_name"] = 'Wolffoni'


        form["situations"] = ['3', '4', '5']  # RelationCheckBoxListWidget

        resp = form.submit("submit")

        dbsession.expire(exc)

        assert exc.hash == re.match(
            r'http://localhost/excavations/(.*)\?msg_col=submit_ok',
            resp.location).group(1)

        assert "New description" == exc.description
        assert "New motif" == exc.motif
        assert 4 == exc.location_district_id  # name="La Claie-aux-Moines"
        assert "rue des écoles" == exc.location_street
        assert "73000" == exc.location_postal_code
        assert "Beçanson" == exc.location_town
        assert 4 == exc.address_id  # label="Lugano"
        assert None == exc.location_position
        assert "mr" == exc.responsible_title
        assert "lapin" == exc.responsible_name
        assert "malin" == exc.responsible_first_name
        assert "06.75.65.83" == exc.responsible_mobile
        assert "lapin.malin@lugano.ch" == exc.responsible_mail
        assert "csc" == exc.responsible_company
        assert "Raoul" == exc.contact_persons[0].first_name
        assert False == exc.validated
        assert None == exc.work_footprint

        situation_ids = [situation.id for situation
                          in exc.situations]
        assert set([3, 4, 5]) == set(situation_ids)
        assert 'Your submission has been taken into account.' == \
            resp.follow().html.find('div', {'class': 'msg-lbl'}).getText()

    def test_edit_new_then_save(self, dbsession, test_app):
        resp = test_app.get("/excavations/new", status=200)

        form = resp.form
        assert '' == form['reference_number'].value
        assert '' == form['hash'].value
        assert '' == form.fields.get('id')[0].value
        assert '' == form['date'].value
        assert '' == form["description"].value
        assert '' == form["motif"].value
        assert form.fields.get('first_name') is None # no contact_persons
        assert '' == form["location_district_id"].value
        assert '' == form["location_street"].value
        assert '' == form["location_postal_code"].value
        assert '' == form["location_town"].value
        assert '' == form["address_id"].value
        assert '' == form["location_position"].value
        assert '' == form["responsible_title"].value
        assert '' == form["responsible_name"].value
        assert '' == form["responsible_first_name"].value
        assert '' == form["responsible_mobile"].value
        assert '' == form["responsible_mail"].value
        assert '' == form["responsible_company"].value
        assert None == form["validated"].value
        assert '' == form["work_footprint"].value
        # assert form["photos"].value == exc.photos

        all_situations = dbsession.query(Situation).order_by(Situation.name).all()
        self.check_checkboxes_list_widget(resp, all_situations, Excavation())

        form["description"] = "Description"
        form["motif"] = "Motif"
        form["location_district_id"] = 4  # name="La Claie-aux-Moines"
        form["location_street"] = "rue des écoles"
        form["location_postal_code"] = '73000'
        form["location_town"] = "Beçanson"
        form["address_id"] = 4  # label="Lugano"
        form["location_position"] = ''
        form["responsible_title"] = "mr"
        form["responsible_name"] = "lapin"
        form["responsible_first_name"] = "malin"
        form["responsible_mobile"] = "06.75.65.83"
        form["responsible_mail"] = "lapin.malin@lugano.ch"
        form["responsible_company"] = "csc"
        form["validated"] = False
        form["work_footprint"] = ''
        form['situations'] = ['3', '4', '5']  # RelationCheckBoxListWidget

        resp = form.submit("submit")

        exc = self._excavation_by_desc(dbsession, "Description")

        assert exc.hash == re.match(
            r'http://localhost/excavations/(.*)\?msg_col=submit_ok',
            resp.location).group(1)

        assert "Description" == exc.description
        assert "Motif" == exc.motif
        assert 4 == exc.location_district_id  # name="La Claie-aux-Moines"
        assert "rue des écoles" == exc.location_street
        assert "73000" == exc.location_postal_code
        assert "Beçanson" == exc.location_town
        assert 4 == exc.address_id  # label="Lugano"
        assert None == exc.location_position
        assert "mr" == exc.responsible_title
        assert "lapin" == exc.responsible_name
        assert "malin" == exc.responsible_first_name
        assert "06.75.65.83" == exc.responsible_mobile
        assert "lapin.malin@lugano.ch" == exc.responsible_mail
        assert "csc" == exc.responsible_company
        assert False == exc.validated
        assert None == exc.work_footprint

        situation_ids = [situation.id for situation in exc.situations]
        assert set([3, 4, 5]) == set(situation_ids)
        assert 'Your submission has been taken into account.' == \
            resp.follow().html.find('div', {'class': 'msg-lbl'}).getText()

    def test_edit_new_then_save_with_missing_field(self, dbsession, test_app):
        resp = test_app.get("/excavations/new", status=200)

        resp = resp.form.submit("submit")

        assert 'There was a problem with your submission' == \
            resp.html.select_one('div[class="error-msg-lbl"]').text
        assert 'Errors have been highlighted below' == \
            resp.html.select_one('div[class="error-msg-detail"]').text
        assert ['Postal Code', 'Street', 'Town'] == \
            sorted([(x.select_one("label").text.strip()) \
            for x in resp.html.select(".has-error")])

    def test_duplicate(self, dbsession, test_app):
        exc = self._excavation_by_ref(dbsession, "ref0001")
        resp = test_app.get("/excavations/{}/duplicate".format(exc.hash), status=200)
        exc = self._excavation_by_ref(dbsession, "ref0001")
        form = resp.form

        assert '' == form['reference_number'].value
        assert '' == form['hash'].value
        assert '' == form.fields.get('id')[0].value
        assert None == form["validated"].value
        assert str(exc.request_date) == form['date'].value
        assert exc.description == form["description"].value
        assert exc.motif == form["motif"].value
        assert str(exc.location_district_id) == form["location_district_id"].value
        assert exc.location_street == form["location_street"].value
        assert exc.location_postal_code == form["location_postal_code"].value
        assert exc.location_town == form["location_town"].value
        assert str(exc.address_id or "") == form["address_id"].value
        assert str(exc.location_position or "") == form["location_position"].value
        assert exc.responsible_title == form["responsible_title"].value
        assert exc.responsible_name == form["responsible_name"].value
        assert exc.responsible_first_name == form["responsible_first_name"].value
        assert exc.responsible_mobile == form["responsible_mobile"].value
        assert exc.responsible_mail == form["responsible_mail"].value
        assert exc.responsible_company == form["responsible_company"].value
        assert str(exc.work_footprint or "") == form["work_footprint"].value
        all_situations = dbsession.query(Situation).order_by(Situation.name).all()
        self.check_checkboxes_list_widget(resp, all_situations, exc)
        assert exc.contact_persons[0].first_name == form["first_name"].value
        assert exc.contact_persons[0].last_name == form["last_name"].value
        assert '' == form.fields.get('id')[1].value
        assert exc.contact_persons[0].id != form.fields.get('id')[1].value
        assert '' == form['permission_id'].value
        assert exc.contact_persons[0].permission_id != form['permission_id'].value

        assert 'Please check that the copy fits before submitting.' == \
            resp.html.find('div', {'class': 'msg-lbl'}).getText()


    def test_unique_validator(self, dbsession, test_app):
        exc = self._excavation_by_ref(dbsession, "ref0001")
        resp = test_app.get("/excavations/{}/duplicate".format(exc.hash), status=200)
        form = resp.form
        form['reference_number'] = exc.reference_number
        resp = form.submit("submit")
        assert 'There was a problem with your submission' == \
            resp.html.select_one('div[class="error-msg-lbl"]').text
        assert 'Errors have been highlighted below' == \
            resp.html.select_one('div[class="error-msg-detail"]').text
        assert '{} is already used.'.format(exc.reference_number) == \
            resp.html.select_one(".has-error").select_one(".help-block").getText()

    def test_delete_success(self, dbsession, test_app):
        exc = self._excavation_by_ref(dbsession, "ref0001")

        resp = test_app.delete("/excavations/{}".format(exc.hash),
            status=200)

        exc = self._excavation_by_ref(dbsession, "ref0001")
        assert None == exc



import configparser
import os

import pyramid.config.settings


def apply_local_settings(settings: pyramid.config.settings.Settings) -> None:
    """Apply to the ``settings`` dict the settings found in the local settings
    file.
    """
    local_settings_path = settings.get("local_settings_path")
    if local_settings_path is not None and os.path.exists(local_settings_path):
        config_parser = configparser.ConfigParser()
        config_parser.read(local_settings_path)
        settings.update(config_parser.items("app:main"))



from sqlalchemy import Column, Integer, LargeBinary, Text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import scoped_session, sessionmaker
from zope.sqlalchemy import register

from .ext import colander_ext

DBSession = scoped_session(sessionmaker())
register(DBSession)
Base = declarative_base()


class FileData:
    id = Column(Integer, primary_key=True)
    filename = Column(Text, nullable=True)
    data = Column(LargeBinary, nullable=False, info={"colanderalchemy": {"typ": colander_ext.BinaryData()}})



from typing import Any, Union

import pyramid.request
import pyramid.response
import sqlalchemy.orm
from pyramid.httpexceptions import HTTPFound
from pyramid.view import view_config

from c2cgeoform import JSONDict, JSONList


@view_config(route_name="c2cgeoform_locale", renderer="json")  # type: ignore[misc]
def set_locale_cookie(request: pyramid.request.Request) -> Union[JSONDict, pyramid.response.Response]:
    """View to change the preferred language."""
    if request.GET["language"]:
        language = request.GET["language"]
        request.response.set_cookie("_LOCALE_", value=language, max_age=31536000)  # max_age = year
    if request.referer is not None:
        return HTTPFound(location=request.referer, headers=request.response.headers)
    return {"success": True}


class ApplicationViewPredicate:
    """
    Predicate which checks request.application.name() match with passed value.

    Example usage

    .. code-block:: python

        @view_defaults(application='admin')
        class AdminViews():
    """

    def __init__(self, application: str, config: Any):
        del config
        self._application = application

    def text(self) -> str:
        return f"application = {self._application}"

    phash = text

    def __call__(self, info: Any, request: pyramid.request.Request) -> bool:
        return request.application.name() == self._application  # type: ignore[no-any-return]


class TableViewPredicate:
    """
    Predicate which checks request.matchdict['table'] match with passed value.

    Example usage

    .. code-block:: python

        @view_defaults(application='admin', table='users')
        class AdminUserViews():
    """

    def __init__(self, table: str, config: Any):
        del config
        self._table = table

    def text(self) -> str:
        return f"table = {self._table}"

    phash = text

    def __call__(self, info: Any, request: pyramid.request.Request) -> bool:
        return info["match"]["table"] == self._table  # type: ignore[no-any-return]


def includeme(config: pyramid.config.Configurator) -> None:
    config.add_view_predicate("application", ApplicationViewPredicate)
    config.add_view_predicate("table", TableViewPredicate)



import os

from c2cgeoform.routes import register_models

from .models.c2cgeoform_demo import Excavation


def includeme(config):
    config.add_static_view("static", "static", cache_max_age=3600)
    config.add_static_view("node_modules", "{{cookiecutter.package}}:node_modules/")
    config.override_asset(
        to_override="{{cookiecutter.package}}:node_modules/",
        override_with=os.path.join(os.path.dirname(__file__), "..", "node_modules"),
    )
    config.add_route("home", "/")
    config.add_route("bus_stops", "/bus_stops")
    config.add_route("addresses", "/addresses")

    register_models(config, [("excavations", Excavation)])



from geoalchemy2.shape import to_shape
from pyramid.view import view_config
from shapely.geometry import mapping

from ..models.c2cgeoform_demo import BusStop


@view_config(route_name="bus_stops", request_method="GET", renderer="json")
def bus_stops(request):
    stops = request.dbsession.query(BusStop).all()
    return {
        "type": "FeatureCollection",
        "features": [
            {
                "type": "Feature",
                "id": stop.id,
                "properties": {"name": stop.name},
                "geometry": mapping(to_shape(stop.geom)),
            }
            for stop in stops
        ],
    }



import pprint
from typing import Any, cast

import deform
import pyramid.response

from c2cgeoform import JSONDict, JSONList


class AbstractViewsTests:
    """
    Base class for views testing
    """

    _prefix = None  # url prefix (index view url). Example : /users

    def get(self, test_app: Any, path: str = "", locale: str = "en", status: int = 200, **kwargs: Any) -> Any:
        return test_app.get(
            f"{self._prefix}{path}",
            headers={"Cookie": f"_LOCALE_={locale}"},
            status=status,
            **kwargs,
        )

    def get_item(self, test_app: Any, item_id: str, **kwargs: Any) -> pyramid.response.Response:
        return self.get(test_app, f"/{item_id}", **kwargs)

    def check_left_menu(self, resp: pyramid.response.Response, title: str) -> None:
        link = resp.html.select_one(".navbar li.active a")
        assert f"http://localhost{self._prefix}" == link.attrs["href"]
        assert title == link.getText()

    def check_grid_headers(
        self, resp: pyramid.response.Response, expected_col_headers: JSONList, check_actions: bool = True
    ) -> None:
        pretty_printer = pprint.PrettyPrinter(indent=4)
        effective_cols = [
            (th.attrs["data-field"], th.getText(), th.attrs["data-sortable"]) for th in resp.html.select("th")
        ]
        expected_col_headers = [[x[0], x[1], len(x) == 3 and x[2] or "true"] for x in expected_col_headers]  # type: ignore[arg-type,index]
        assert expected_col_headers == effective_cols, str.format(  # type: ignore[comparison-overlap]
            "\n\n{}\n\n differs from \n\n{}",
            pretty_printer.pformat(expected_col_headers),
            pretty_printer.pformat(effective_cols),
        )
        if check_actions:
            actions = resp.html.select_one('th[data-field="actions"]')
            assert "false" == actions.attrs["data-sortable"]

    def check_new_button(self, resp: pyramid.response.Response) -> None:
        assert 1 == len(list(filter(lambda x: str(x.contents) == "['New']", resp.html.findAll("a"))))

    def check_search(
        self,
        test_app: Any,
        search: str = "",
        offset: int = 0,
        limit: int = 10,
        sort: str = "",
        order: str = "",
        total: Any = None,
        **kwargs: Any,
    ) -> Any:
        json = test_app.get(
            f"{self._prefix}/grid.json",
            params={
                "offset": offset,
                "limit": limit,
                "search": search,
                "sort": sort,
                "order": order,
                **kwargs,
            },
            status=200,
        ).json
        if total is not None:
            assert total == json["total"]
        return json

    def check_checkboxes(self, form: deform.Form, name: str, expected: list[JSONDict]) -> None:
        for i, exp in enumerate(expected):
            field = form.get(name, index=i)
            checkbox = form.html.select_one(f"#{field.id}")
            label = form.html.select_one(f"label[for={field.id}]")
            assert exp["label"] == list(label.stripped_strings)[0]
            assert exp["value"] == checkbox["value"]
            assert exp["checked"] == field.checked

    def get_first_field_named(self, form: deform.Form, name: str) -> Any:
        return form.fields.get(name)[0]

    def set_first_field_named(self, form: deform.Form, name: str, value: Any) -> None:
        form.fields.get(name)[0].value__set(value)

    def _check_sequence(self, sequence: Any, expected: Any) -> None:
        seq_items = sequence.select(".deform-seq-item")
        assert len(expected) == len(seq_items)
        for seq_item, exp in zip(seq_items, expected):
            self._check_mapping(seq_item, exp)

    def _check_mapping(self, mapping_item: Any, expected: list[JSONDict]) -> None:
        for exp in expected:
            input_tag = mapping_item.select_one('[name="{}"]'.format(exp["name"]))
            if "value" in exp:
                if exp.get("readonly", False):
                    item = mapping_item.select_one(".item-{}".format(exp["name"]))
                    assert input_tag is None
                    assert exp["value"] or "" == item.select("p").stripped_strings[0]
                elif input_tag.name == "select":
                    value = exp["value"]
                    assert isinstance(value, list)
                    self._check_select(input_tag, cast(list[JSONDict], value))
                elif input_tag.name == "textarea":
                    assert (exp["value"] or "") == (input_tag.string or "")
                else:
                    assert (exp["value"] or "") == input_tag.attrs.get("value", "")
            if exp.get("hidden", False):
                assert "hidden" == input_tag["type"]
            if "label" in exp:
                label_tag = mapping_item.select_one('label[for="{}"]'.format(input_tag["id"]))
                assert exp["label"] == label_tag.getText().strip()

    def _check_select(self, select: Any, expected: list[JSONDict]) -> None:
        for exp, option in zip(expected, select.find_all("option")):
            if "text" in exp:
                assert exp["text"] == option.text
            if "value" in exp:
                assert exp["value"] == option["value"]
            if "selected" in exp:
                assert exp["selected"] == ("selected" in option.attrs)

    def _check_submission_problem(self, resp: pyramid.response.Response, expected_msg: str) -> None:
        assert (
            "There was a problem with your submission"
            == resp.html.select_one('div[class="error-msg-lbl"]').text
        )
        assert (
            "Errors have been highlighted below" == resp.html.select_one('div[class="error-msg-detail"]').text
        )
        assert (
            expected_msg
            == resp.html.select_one("[class~='has-error']")
            .select_one("[class~='help-block']")
            .getText()
            .strip()
        )



import unittest
import unittest.mock as mock
from unittest.mock import patch

import colander
from sqlalchemy import Column, Integer, String

from c2cgeoform.models import Base
from c2cgeoform.schema import GeoFormSchemaNode


class FieldsCollection(Base):
    __tablename__ = "many_types_and_contraints_fields"

    id = Column(Integer, primary_key=True)
    text = Column(String(length=4))


class TestUniqueValidator(unittest.TestCase):
    def test_constraint_on_column_at_sql_alchemy_side_throw_colander_invalid(self):
        schema_node = GeoFormSchemaNode(FieldsCollection)
        schema_node.deserialize({"text": "foo"})
        try:
            schema_node.deserialize({"text": "more than five char"})
            assert False
        except colander.Invalid as e:
            assert "Longer than maximum length" in e.children[0].msg

    def test_custom_validator_triggered_at_serialization(self):
        schema_node = GeoFormSchemaNode(FieldsCollection)

        def side_effect(node, value):
            raise colander.Invalid(schema_node["text"], msg="test ERROR !")

        schema_node["text"].validator = side_effect
        try:
            schema_node.deserialize({"text": "foo"})
            assert False
        except colander.Invalid as e:
            assert "test ERROR !" == e.children[0].msg

    @patch("c2cgeoform.schema.unique_validator")
    def test_adding_validator_does_not_overrides_sqlalchemy_ones(self, unique_validator_mock):
        schema_node = GeoFormSchemaNode(FieldsCollection)
        schema_node.add_unique_validator(FieldsCollection.text, FieldsCollection.id)
        try:
            schema_node.deserialize({"text": "more than five char"})
            assert False
        except colander.Invalid as e:
            assert "Longer than maximum length" in e.children[0].msg[0]
            unique_validator_mock.assert_has_calls(
                [mock.call(mock.ANY, mock.ANY, mock.ANY, mock.ANY, "more than five char")]
            )

    @patch("c2cgeoform.schema.unique_validator")
    def test_adding_validator_does_not_overrides_custom_ones(self, unique_validator_mock):
        schema_node = GeoFormSchemaNode(FieldsCollection)

        def side_effect(node, value):
            raise colander.Invalid(schema_node["text"], msg="test ERROR !")

        schema_node["text"].validator = side_effect
        schema_node.add_unique_validator(FieldsCollection.text, FieldsCollection.id)
        try:
            schema_node.deserialize({"text": "foo"})
            assert False
        except colander.Invalid as e:
            assert "test ERROR !" == e.children[0].msg[0]
            unique_validator_mock.assert_has_calls([mock.call(mock.ANY, mock.ANY, mock.ANY, mock.ANY, "foo")])



from functools import partial
from typing import Any, Optional, cast

import colander
import pyramid.request
import sqlalchemy.orm
import sqlalchemy.sql.elements
from colanderalchemy import SQLAlchemySchemaNode
from sqlalchemy import and_, or_
from sqlalchemy.inspection import inspect

from c2cgeoform import JSONDict, _


@colander.deferred  # type: ignore[misc]
def deferred_request(node: Any, kwargs: dict[str, Any]) -> pyramid.request.Request:
    del node  # unused
    return kwargs.get("request")


@colander.deferred  # type: ignore[misc]
def deferred_dbsession(node: Any, kwargs: dict[str, Any]) -> sqlalchemy.orm.Session:
    del node  # unused
    return cast(sqlalchemy.orm.Session, kwargs.get("dbsession"))


def unique_validator(
    mapper: type[Any],
    column: sqlalchemy.orm.attributes.InstrumentedAttribute[Any],
    id_column: sqlalchemy.orm.attributes.InstrumentedAttribute[Any],
    node: str,
    value: Any,
) -> None:
    dbsession: sqlalchemy.orm.Session = node.bindings["dbsession"]  # type: ignore[attr-defined]
    _id = node.bindings["request"].matchdict["id"]  # type: ignore[attr-defined]
    _id = _id if _id != "new" else None
    if dbsession.query(mapper).filter(column == value, id_column != _id).count() != 0:
        raise colander.Invalid(node, _("{} is already used.").format(value))


class GeoFormSchemaNode(SQLAlchemySchemaNode):  # type: ignore[misc] # pylint: disable=abstract-method
    """
    An SQLAlchemySchemaNode with deferred request and dbsession properties.
    This will allow defining schemas that requires the request and dbsession at
    module-scope.

    Example usage:

    .. code-block:: python

        schema = GeoFormSchemaNode(MyModel)

        def create_form(request, dbsession):
            return Form(
                schema = schema.bind(
                    request=request,
                    dbsession=request.dbsession),
                ...
            )
    """

    def __init__(self, *args: Any, **kw: Any) -> None:
        super().__init__(*args, **kw)
        self.request = deferred_request
        self.dbsession = deferred_dbsession

    def add_unique_validator(
        self,
        column: sqlalchemy.orm.attributes.InstrumentedAttribute[Any],
        column_id: sqlalchemy.orm.attributes.InstrumentedAttribute[Any],
    ) -> None:
        """
        Adds an unique validator on this schema instance.

        column
            SQLAlchemy ColumnProperty that should be unique.

        column_id
            SQLAlchemy MapperProperty that is used to recognize the entity,
            basically the primary key ColumnProperty.
        """
        validator = partial(unique_validator, self.class_, column, column_id)
        if self[column.name].validator is None:
            self[column.name].validator = validator
        else:
            self[column.name].validator = colander.All(self[column.name].validator, validator)


class GeoFormManyToManySchemaNode(GeoFormSchemaNode):  # pylint: disable=abstract-method
    """
    A GeoFormSchemaNode that properly handles many to many relationships.

    includes:
        Default to primary key name(s) only.
    """

    def __init__(self, class_: type[Any], includes: Optional[list[str]], *args: Any, **kw: Any) -> None:
        includes = [pk.name for pk in inspect(class_).primary_key]
        super().__init__(class_, includes, *args, **kw)

    def objectify(self, dict_: JSONDict, context: Any = None) -> Any:
        """
        Method override that returns the existing ORM class instance instead of
        creating a new one.
        """
        dbsession = self.bindings["dbsession"]
        class_ = self.inspector.class_
        return dbsession.query(class_).get(dict_.values())


def manytomany_validator(node: type[Any], cstruct: list[JSONDict]) -> None:
    """
    Validator function that checks if ``cstruct`` values exist in the related table.

    Note that entities are retrieved using only one query and placed in
    SQLAlchemy identity map before looping on ``cstruct``.
    """
    dbsession = node.bindings["dbsession"]
    class_ = node.children[0].inspector.class_
    query = dbsession.query(class_)
    filters = []
    for dict_ in cstruct:
        filters.append(and_(*[getattr(class_, key) == value for key, value in dict_.items()]))
    query = query.filter(or_(*filters))
    results = query.all()  # get all records in cache in one request
    diff = {tuple(dict_.values()) for dict_ in cstruct} - {inspect(a).identity for a in results}
    if len(diff) > 0:
        raise colander.Invalid(
            "Values {} does not exist in table {}".format(
                ", ".join(str(identity) for identity in diff), class_.__tablename__
            )
        )



import io
import json
import os
from io import BytesIO
from typing import Any, Union

import colander
import pyproj
from colander import Invalid, SchemaType
from geoalchemy2 import WKBElement
from geoalchemy2.shape import from_shape, to_shape
from shapely.geometry import mapping, shape
from shapely.ops import transform
from typing_extensions import Buffer


class Geometry(SchemaType):  # type: ignore[misc]
    """A Colander type meant to be used with GeoAlchemy 2 geometry columns.

    Example usage

    .. code-block:: python

        geom = Column(
            geoalchemy2.Geometry('POLYGON', 4326), info={
                'colanderalchemy': {
                    'typ': colander_ext.Geometry(
                        'POLYGON', srid=4326, map_srid=3857),
                    'widget': deform_ext.MapWidget()
                }})

    **Attributes/Arguments**

    geometry_type
        The geometry type should match the column geometry type.

    srid
        The SRID of the geometry should also match the column definition.

    map_srid
        The projection used for the OpenLayers map. The geometries will be
        reprojected to this projection.

    """

    def __init__(self, geometry_type: str = "GEOMETRY", srid: int = -1, map_srid: int = -1) -> None:
        self.geometry_type = geometry_type.upper()
        self.srid = int(srid)
        self.map_srid = int(map_srid)
        if self.map_srid == -1:
            self.map_srid = self.srid

        if self.srid != self.map_srid:
            self.project_db_to_map = pyproj.Transformer.from_crs(
                self.srid,
                self.map_srid,
                always_xy=True,
            ).transform

            self.project_map_to_db = pyproj.Transformer.from_crs(
                self.map_srid,
                self.srid,
                always_xy=True,
            ).transform

    def serialize(
        self, node: Any, appstruct: Union[colander._null, WKBElement]
    ) -> Union[colander._null, str]:
        """
        In Colander speak: Converts a Python data structure (an appstruct) into
        a serialization (a cstruct).
        Or: Converts a `WKBElement` into a GeoJSON string.
        """

        if appstruct is colander.null:
            return colander.null
        if isinstance(appstruct, WKBElement):
            geometry = to_shape(appstruct)
            if self.map_srid not in (self.srid, appstruct.srid):
                geometry = transform(self.project_db_to_map, geometry)

            return json.dumps(mapping(geometry))
        raise Invalid(node, "Unexpected value: %r" % appstruct)

    def deserialize(self, node: Any, cstruct: Union[colander._null, str]) -> Union[colander._null, str]:
        """
        In Colander speak: Converts a serialized value (a cstruct) into a
        Python data structure (a appstruct).
        Or: Converts a GeoJSON string into a `WKBElement`.
        """

        if cstruct is colander.null or cstruct == "":
            return colander.null
        try:
            # TODO Shapely does not support loading GeometryCollections from
            # GeoJSON, see https://github.com/Toblerity/Shapely/issues/115
            geometry = shape(json.loads(cstruct))
        except Exception as exception:
            raise Invalid(node, f"Invalid geometry: {cstruct!r}") from exception

        if self.srid != self.map_srid:
            geometry = transform(self.project_map_to_db, geometry)

        return from_shape(geometry, srid=self.srid)


class BinaryData(SchemaType):  # type: ignore[misc]
    """A Colander type meant to be used with ``LargeBinary`` columns.

    Example usage

    .. code-block:: python

        class Model():
            id = Column(Integer, primary_key=True)
            data = Column(LargeBinary, info={
                'colanderalchemy': {
                    'typ': colander_ext.BinaryData()
                }})

    It is usually not used directly in application models, but through
    the ``c2cgeoform.models.FileData`` mixin, which is meant to be used
    with a ``deform_ext.FileUploadWidget``.

    The serialize method just returns ``colander.null``. This is because
    the ``FileUploadWidget``'s template does not use and need the binary
    data.

    The deserialize method gets a Python ``file`` object and returns a
    bytes string that is appropriate for the database.

    """

    def serialize(
        self, node: colander.SchemaNode, appstruct: Union[colander._null, Buffer]
    ) -> Union[colander._null, io.BytesIO]:
        """
        In Colander speak: Converts a Python data structure (an appstruct) into
        a serialization (a cstruct).
        """
        del node  # unused

        if appstruct is colander.null or appstruct == "":
            return colander.null
        return BytesIO(appstruct)

    def deserialize(
        self, node: colander.SchemaNode, cstruct: Union[colander._null, str, io.IOBase]
    ) -> Union[colander._null, bytes]:
        """
        In Colander speak: Converts a serialized value (a cstruct) into a
        Python data structure (a appstruct).
        Or: Converts a Python file stream to plain binary data.
        """
        del node  # unused

        if cstruct is colander.null or cstruct == "":
            return colander.null
        assert isinstance(cstruct, io.IOBase)
        byte_array = cstruct.read()
        # set the file position back to 0, so that the file can be read again
        cstruct.seek(0, os.SEEK_SET)
        return byte_array



import os
import time
import unittest

from pyramid import testing
from pyramid.paster import get_appsettings
from sqlalchemy import engine_from_config, text
from webob.multidict import MultiDict

from c2cgeoform import init_deform
from c2cgeoform.models import Base, DBSession
from c2cgeoform.settings import apply_local_settings


def wait_for_db(engine):
    sleep_time = 1
    max_sleep = 30
    while sleep_time < max_sleep:
        try:
            with engine.connect() as connection:
                connection.execute(text("SELECT 1;"))
            return
        except Exception as e:
            print(str(e))
            print("Waiting for the DataBase server to be reachable")
            time.sleep(sleep_time)
            sleep_time *= 2
    exit(1)  # noqa


class DatabaseTestCase(unittest.TestCase):
    def setUp(self):  # noqa
        curdir = os.path.dirname(os.path.abspath(__file__))
        configfile = os.path.realpath(os.path.join(curdir, "../../tests.ini"))
        settings = get_appsettings(configfile)
        apply_local_settings(settings)
        engine = engine_from_config(settings, "sqlalchemy.")
        DBSession.configure(bind=engine)

        wait_for_db(engine)

        from .models_test import EmploymentStatus, Person, Tag  # noqa

        Base.metadata.create_all(engine)
        self.cleanup()

        # fill some test data into the `EmploymentStatus` and `Tags` table
        DBSession.add(EmploymentStatus(id=0, name="Worker"))
        DBSession.add(EmploymentStatus(id=1, name="Employee"))
        DBSession.add(EmploymentStatus(id=2, name="Self-employed and contractor"))
        DBSession.add(EmploymentStatus(id=3, name="Director"))
        DBSession.add(EmploymentStatus(id=4, name="Office holder"))

        DBSession.add(Tag(id=0, name="Tag A"))
        DBSession.add(Tag(id=1, name="Tag B"))
        DBSession.add(Tag(id=2, name="Tag C"))
        DBSession.add(Tag(id=3, name="Tag D"))
        DBSession.add(Tag(id=4, name="Tag E"))

        self.request = testing.DummyRequest(post=MultiDict(), dbsession=DBSession)
        testing.setUp(request=self.request)

        config = testing.setUp()
        config.add_route("form", "/{schema}/form/")
        config.add_route("view_user", "/{schema}/form/{hash}")
        config.add_route("confirm", "/{schema}/form/confirm")
        init_deform("c2cgeoform")

    def tearDown(self):  # noqa
        self.cleanup()
        DBSession.remove()
        testing.tearDown()

    def cleanup(self):
        from .models_test import EmploymentStatus, Person, Phone, Tag

        DBSession.query(Tag).delete()
        DBSession.query(Phone).delete()
        DBSession.query(Person).delete()
        DBSession.query(EmploymentStatus).delete()



import os

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, "README.txt")) as f:
    README = f.read()

setup(
    name="{{cookiecutter.project}}",
    version="0.0",
    description="{{cookiecutter.project}}",
    long_description=README,
    classifiers=[
        "Programming Language :: Python",
        "Framework :: Pyramid",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
    ],
    author="",
    author_email="",
    url="",
    keywords="web wsgi bfg pylons pyramid",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    test_suite="{{cookiecutter.package}}",
    entry_points="""\
      [paste.app_factory]
      main = {{cookiecutter.package}}:main
      [console_scripts]
      initialize_{{cookiecutter.package}}_db = {{cookiecutter.package}}.scripts.initializedb:main
      """,
)



from pyramid.httpexceptions import HTTPBadRequest
from pyramid.view import view_config

from ..models.c2cgeoform_demo import Address


@view_config(route_name="addresses", request_method="GET", renderer="json")
def addresses(request):
    if "term" not in request.params:
        return HTTPBadRequest()
    term = "%%%s%%" % request.params["term"]
    query = request.dbsession.query(Address).filter(Address.label.ilike(term))
    return [{"id": addr.id, "label": addr.label} for addr in query]



import logging
from typing import Any, Callable, Dict, Generic, Optional, Tuple, TypedDict, TypeVar, Union, cast

import geojson
import pyramid.request
import pyramid.response
import sqlalchemy.engine.row
import sqlalchemy.orm.attributes
import sqlalchemy.orm.properties
import sqlalchemy.schema
import sqlalchemy.sql.elements
import sqlalchemy.sql.expression
from deform import Form, ValidationFailure  # , ZPTRendererFactory
from deform.form import Button
from geoalchemy2.elements import WKBElement
from geoalchemy2.shape import to_shape
from geojson import Feature, FeatureCollection
from pyramid.httpexceptions import HTTPFound, HTTPInternalServerError, HTTPNotFound
from sqlalchemy import desc, or_, types
from sqlalchemy.exc import DBAPIError
from sqlalchemy.inspection import inspect
from sqlalchemy.orm.properties import ColumnProperty
from sqlalchemy.orm.relationships import RelationshipProperty
from translationstring import TranslationString

from c2cgeoform import JSON, JSONDict, JSONList, _, default_map_settings

_LOGGER = logging.getLogger(__name__)

_DB_ERR_MSG = """\
Pyramid is having a problem using your SQL database.  The problem
might be caused by one of the following things:

1.  You may need to run the "initialize_c2cgeoportal_admin_db" script
    to initialize your database tables.  Check your virtual
    environment's "bin" directory for this script and try to run it.

2.  Your database server may not be running.  Check that the
    database server referred to by the "sqlalchemy.url" setting in
    your "development.ini" file is running.

After you fix the problem, please restart the Pyramid application to
try it again.
"""


def model_attr_info(
    attr: Optional[
        Union[
            str,
            sqlalchemy.orm.attributes.InstrumentedAttribute[Any],
            sqlalchemy.orm.relationships.Relationship[Any],
            sqlalchemy.orm.relationships.RelationshipProperty[Any],
            sqlalchemy.sql.schema.Column[Any],
            sqlalchemy.sql.elements.NamedColumn[Any],
        ]
    ],
    *keys: str,
    default: Any = None,
) -> Any:
    if attr is None:
        return default

    assert isinstance(
        attr,
        (
            sqlalchemy.orm.attributes.InstrumentedAttribute,
            sqlalchemy.orm.relationships.Relationship,
            sqlalchemy.orm.relationships.RelationshipProperty,
            sqlalchemy.sql.schema.Column,
            sqlalchemy.sql.elements.NamedColumn,
        ),
    ), type(attr)
    value = attr.info
    for key in keys:
        if key not in value:
            return default
        value = value[key]
    return value


T = TypeVar("T", bound=type)


def _getattr(
    model: Optional[type[T]], attr: Optional[Union[sqlalchemy.schema.Column[Any], str]]
) -> sqlalchemy.schema.Column[Any]:
    if model is None:
        assert isinstance(attr, sqlalchemy.schema.Column)
        return attr
    assert isinstance(attr, str)
    return cast(sqlalchemy.schema.Column[Any], getattr(model, attr))


class ListField(Generic[T]):
    def __init__(
        self,
        model: Optional[type[T]] = None,
        attr: Optional[str] = None,
        key: Optional[str] = None,
        label: Optional[str] = None,
        renderer: Optional[Callable[[T], JSONDict]] = None,
        sort_column: Optional[sqlalchemy.sql.elements.ColumnElement[Any]] = None,
        filter_column: Optional[sqlalchemy.sql.elements.ColumnElement[Any]] = None,
        visible: bool = True,
    ):
        self._attr = _getattr(model, attr)
        self._key = key or self._attr.key
        self._label = label or model_attr_info(self._attr, "colanderalchemy", "title") or self._key
        self._renderer = renderer or self._prop_renderer
        is_column = isinstance(self._attr.property, ColumnProperty)
        self._sort_column = sort_column or (self._attr if is_column else None)
        self._filter_column = (
            filter_column if filter_column is not None else self._attr if is_column else None
        )
        self._visible = visible

    def _prop_renderer(self, entity: T) -> str:
        value = None
        if self._attr is not None:
            value = getattr(entity, self._attr.key)
        if value is None:
            return ""
        if isinstance(value, WKBElement):
            return "Geometry"
        return str(value)

    def id(self) -> Optional[str]:  # pylint: disable=invalid-name
        return self._key

    def label(self) -> str:
        return self._label

    def value(self, entity: T) -> JSON:
        return self._renderer(entity)

    def sortable(self) -> bool:
        return self._sort_column is not None

    def filtrable(self) -> bool:
        return self._filter_column is not None and isinstance(self._filter_column.type, types.String)

    def sort_column(self) -> sqlalchemy.sql.expression.ColumnElement[Any]:
        assert self._sort_column is not None
        return self._sort_column

    def filter_expression(self, term: str) -> sqlalchemy.sql.expression.BinaryExpression[bool]:
        assert self._filter_column is not None
        return self._filter_column.ilike(term)

    def visible(self) -> bool:
        return self._visible


class ItemAction:
    def __init__(
        self,
        name: str,
        url: str,
        method: Union[bool, str] = False,
        label: Optional[str] = None,
        css_class: str = "",
        icon: Optional[str] = None,
        confirmation: str = "",
    ) -> None:
        self._name = name
        self._url = url
        self._method = method
        self._label = label or self._name
        self._css_class = css_class
        self._icon = icon
        self._confirmation = confirmation

    def name(self) -> str:
        return self._name

    def url(self) -> str:
        return self._url

    def method(self) -> Union[bool, str]:
        return self._method

    def label(self) -> str:
        return self._label

    def css_class(self) -> str:
        return self._css_class

    def icon(self) -> Optional[str]:
        return self._icon

    def confirmation(self) -> str:
        return self._confirmation

    def to_dict(self, request: pyramid.request.Request) -> JSONDict:
        return {
            "name": self._name,
            "url": self._url,
            "method": self._method,
            "label": request.localizer.translate(self._label),
            "css_class": self._css_class,
            "icon": self._icon,
            "confirmation": self._confirmation,
        }


class UserMessage:
    def __init__(self, text: str, css_class: str = "alert-success") -> None:
        self._text = text
        self._css_class = css_class

    def text(self) -> str:
        return self._text

    def css_class(self) -> str:
        return self._css_class

    def __str__(self) -> str:
        # For compatibility with old templates
        return self._text


# TODO for Python 3.11
# class IndexResponse(TypedDict, Generic[T]):
#     grid_actions: list[ItemAction]
#     list_fields: list[ListField[T]]
class IndexResponse(TypedDict):
    grid_actions: list[ItemAction]
    list_fields: list[ListField[Any]]


class GridResponse(TypedDict):
    rows: JSONList
    total: int


class MapResponse(TypedDict):
    map_options: Dict[str, Any]


class DeformDependencies:
    css: Dict[str, str]
    js: Dict[str, str]


class ObjectResponse(TypedDict):
    title: str
    form: Form
    form_render_args: Union[tuple[Any], list[Any]]
    form_render_kwargs: dict[str, Any]
    deform_dependencies: DeformDependencies


SaveResponse = Union[HTTPFound, ObjectResponse]


class DeleteResponse(TypedDict):
    success: bool
    redirect: str


class AbstractViews(Generic[T]):
    _model: Optional[type[T]] = None  # sqlalchemy model
    _list_fields: list[ListField[T]] = []  # Fields in list
    _list_ordered_fields: list[
        Union[sqlalchemy.sql.elements.ColumnClause[Any], sqlalchemy.sql.elements.ColumnElement[Any]]
    ] = []  # Fields in list used for default orderby
    _id_field: Optional[str] = None  # Primary key
    _geometry_field: Optional[str] = None  # Geometry field
    _base_schema: Optional[type[T]] = None  # base colander schema

    MSG_COL = {
        "submit_ok": UserMessage(_("Your submission has been taken into account."), "alert-success"),
        "copy_ok": UserMessage(_("Please check that the copy fits before submitting."), "alert-info"),
    }

    def __init__(self, request: pyramid.request.Request) -> None:
        self._request = request
        self._schema: Optional[str] = None
        self._appstruct: Optional[Dict[str, Any]] = None
        self._obj: Optional[T] = None

    def index(self) -> IndexResponse:
        return {
            "grid_actions": self._grid_actions(),
            "list_fields": self._list_fields,
        }

    def grid(self) -> GridResponse:
        """
        API method which serves the JSON data for the Bootgrid table in the admin view.
        """
        try:
            params = self._request.params
            offset = int(params.get("offset", 0) if params.get("offset") != "NaN" else 0)
            limit = int(params.get("limit", -1) if params.get("limit") != "NaN" else -1)
            search = params.get("search", "").strip()
            sort = params.get("sort", "")
            order = params.get("order", "")

            query = self._base_query()
            query = self._filter_query(query, search)
            query = self._sort_query(query, sort, order)

            return {"rows": self._grid_rows(query, offset, limit), "total": query.count()}
        except DBAPIError as exception:
            _LOGGER.error(str(exception), exc_info=True)
            raise HTTPInternalServerError(_DB_ERR_MSG) from exception

    def map(self, map_settings: Optional[JSONDict] = None) -> MapResponse:
        map_settings = map_settings or {}
        map_options = {
            **default_map_settings,
            **{
                "url": self._request.route_url(
                    "c2cgeoform_geojson",
                    _query={
                        "srid": map_settings.get("srid", default_map_settings["srid"]),
                    },
                ),
            },
            **map_settings,
        }
        return {
            "map_options": {
                key: (self._request.translate(value) if isinstance(value, TranslationString) else value)
                for key, value in map_options.items()
            }
        }

    def geojson(self) -> geojson.FeatureCollection:
        assert self._geometry_field is not None
        assert self._id_field is not None

        srid = int(self._request.params.get("srid", 3857))

        query = self._base_query().add_column(
            getattr(self._model, self._geometry_field).ST_Transform(srid).label("_geometry")
        )

        features: list[geojson.Feature] = []
        for entities in query:
            entity = entities[0]
            geometry = entities[-1]
            features.append(
                Feature(
                    id=getattr(entity, self._id_field),
                    geometry=to_shape(geometry) if geometry is not None else None,
                    properties={f.id(): f.value(entity) for f in self._list_fields},
                )
            )
        return FeatureCollection(features)

    def _base_query(self) -> sqlalchemy.orm.query.Query[T]:
        return cast(sqlalchemy.orm.query.Query[T], self._request.dbsession.query(self._model))

    def _filter_query(
        self, query: sqlalchemy.orm.query.Query[T], search_phrase: str
    ) -> sqlalchemy.orm.query.Query[T]:
        if search_phrase != "":
            search_expr = "%" + "%".join(search_phrase.split()) + "%"

            # create `ilike` filters for every list text field
            filters = []
            for field in self._list_fields:
                if field.filtrable():
                    filters.append(field.filter_expression(search_expr))

            # then join the filters into one `or` condition
            if len(filters) > 0:
                query = query.filter(or_(*filters))

        return query

    def _sort_query(
        self, query: sqlalchemy.orm.query.Query[T], sort: str, order: str
    ) -> sqlalchemy.orm.query.Query[T]:
        for field in self._list_fields:
            if field.id() == sort:
                if order == "desc":
                    query = query.order_by(desc(field.sort_column()))
                else:
                    query = query.order_by(field.sort_column())
        # default order by
        for order_field in self._list_ordered_fields:
            query = query.order_by(order_field)
        return query

    def _grid_rows(self, query: sqlalchemy.orm.query.Query[T], offset: int, limit: int) -> JSONList:
        # Sort on primary key as subqueryload with limit need deterministic order
        for pkey_column in inspect(self._model).primary_key:  # type: ignore[union-attr]
            query = query.order_by(pkey_column)

        if limit != -1:
            query = query.limit(limit).offset(offset)
        rows: JSONList = []

        for entities in query:
            # For some reason like:
            # https://docs.sqlalchemy.org/en/14/changelog/migration_20.html#using-distinct-with-additional-columns-but-only-select-the-entity
            # we are required to add a second field to the query, in this case we should get only the first one
            entity: T = (
                cast(sqlalchemy.engine.row.Row[Tuple[T, Any]], entities)[0]
                if isinstance(entities, sqlalchemy.engine.row.Row)
                else entities
            )
            row = cast(
                JSONDict,
                {
                    f.id(): f.value(entity)
                    for f in (self._list_fields + [ListField(self._model, self._id_field, key="_id_")])
                },
            )
            row["actions"] = self._grid_item_actions(entity)
            rows.append(row)
        return rows

    def _form(self, schema: Optional[type[T]] = None, **kwargs: Any) -> Form:
        schema = schema or self._base_schema
        assert schema is not None
        self._schema = schema.bind(request=self._request, dbsession=self._request.dbsession)  # type: ignore[attr-defined]

        form = Form(self._schema, buttons=[Button(name="formsubmit", title=_("Submit"))], **kwargs)

        return form

    def _populate_widgets(self, node: Any) -> None:
        """Populate ``deform_ext.RelationSelectMixin`` widgets."""

        if hasattr(node.widget, "populate"):
            node.widget.populate(self._request.dbsession, self._request)

        for child in node:
            self._populate_widgets(child)

    def _is_new(self) -> bool:
        return self._request.matchdict.get("id") == "new"  # type: ignore[no-any-return]

    def _get_object(self) -> T:
        assert self._id_field is not None

        if self._is_new():
            assert self._model is not None
            return cast(T, self._model())  # type: ignore[call-overload] # pylint: disable=not-callable
        primary_key = self._request.matchdict.get("id")
        obj = (
            self._request.dbsession.query(self._model)
            .filter(getattr(self._model, self._id_field) == primary_key)
            .one_or_none()
        )
        if obj is None:
            raise HTTPNotFound()
        return cast(T, obj)

    def _model_config(self) -> JSONDict:
        return getattr(inspect(self._model).class_, "__c2cgeoform_config__", {})  # type: ignore[union-attr]

    def _grid_actions(self) -> list[ItemAction]:
        return [
            ItemAction(
                name="new",
                label=_("New"),
                css_class="btn btn-primary btn-new",
                url=self._request.route_url("c2cgeoform_item", id="new"),
            )
        ]

    def _grid_item_actions(self, item: T) -> JSONDict:
        assert self._id_field is not None

        actions = self._item_actions(item)
        actions.insert(
            0,
            ItemAction(
                name="edit",
                label=_("Edit"),
                icon="glyphicon glyphicon-pencil",
                url=self._request.route_url("c2cgeoform_item", id=getattr(item, self._id_field)),
            ),
        )
        return {
            "dropdown": [action.to_dict(self._request) for action in actions],
            "dblclick": self._request.route_url("c2cgeoform_item", id=getattr(item, self._id_field)),
        }

    def _item_actions(self, item: T, readonly: bool = False) -> list[ItemAction]:
        assert self._id_field is not None

        actions = []

        inspected_item: Any = inspect(item)
        assert isinstance(inspected_item, sqlalchemy.orm.InstanceState)
        if inspected_item.persistent and self._model_config().get("duplicate", False):
            actions.append(
                ItemAction(
                    name="duplicate",
                    label=_("Duplicate"),
                    icon="glyphicon glyphicon-duplicate",
                    url=self._request.route_url(
                        "c2cgeoform_item_duplicate", id=getattr(item, self._id_field)
                    ),
                )
            )

        if inspected_item.persistent and not readonly:
            actions.append(
                ItemAction(
                    name="delete",
                    label=_("Delete"),
                    icon="glyphicon glyphicon-remove",
                    url=self._request.route_url("c2cgeoform_item", id=getattr(item, self._id_field)),
                    method="DELETE",
                    confirmation=_("Are your sure you want to delete this record ?"),
                )
            )

        return actions

    def edit(self, schema: Optional[type[T]] = None, readonly: bool = False) -> ObjectResponse:
        obj = self._get_object()
        form = self._form(schema=schema, readonly=readonly)
        self._populate_widgets(form.schema)
        dict_ = form.schema.dictify(obj)
        if self._is_new():
            dict_.update(self._request.GET)
        kwargs = {
            "request": self._request,
            "actions": self._item_actions(obj, readonly=readonly),
            "readonly": readonly,
            "obj": obj,
        }
        if "msg_col" in self._request.params and self._request.params["msg_col"] in self.MSG_COL:
            msg = self.MSG_COL[self._request.params["msg_col"]]
            if isinstance(msg, str):
                # For compatibility with old views
                msg = UserMessage(msg, "alert-success")
            kwargs.update({"msg_col": [msg]})
        return {
            "title": form.title,
            "form": form,
            "form_render_args": (dict_,),
            "form_render_kwargs": kwargs,
            "deform_dependencies": form.get_widget_resources(),
        }

    def copy_members_if_duplicates(self, source: T, excludes: Optional[list[str]] = None) -> T:
        dest = cast(T, source.__class__())  # type: ignore[call-overload]
        insp = inspect(source.__class__)

        for prop in insp.attrs:  # type: ignore[union-attr]
            if isinstance(prop, ColumnProperty):
                is_primary_key = prop.columns[0].primary_key
                to_duplicate = model_attr_info(prop.columns[0], "c2cgeoform", "duplicate", default=True)
                to_exclude = excludes and prop.columns[0].key in excludes
                if not is_primary_key and to_duplicate and not to_exclude:
                    setattr(dest, prop.key, getattr(source, prop.key))
            if isinstance(prop, RelationshipProperty):
                if model_attr_info(prop, "c2cgeoform", "duplicate", default=True):
                    if prop.cascade.delete:
                        if not prop.uselist:
                            duplicate: Union[T, list[T]] = self.copy_members_if_duplicates(
                                getattr(source, prop.key)
                            )
                        else:
                            duplicate = [
                                self.copy_members_if_duplicates(m) for m in getattr(source, prop.key)
                            ]
                    else:
                        duplicate = getattr(source, prop.key)
                    setattr(dest, prop.key, duplicate)
        return dest

    def copy(self, src: T, excludes: Optional[list[str]] = None) -> ObjectResponse:
        # excludes only apply at first level
        form = self._form(action=self._request.route_url("c2cgeoform_item", id="new"))
        with self._request.dbsession.no_autoflush:
            dest = self.copy_members_if_duplicates(src, excludes)
            dict_ = form.schema.dictify(dest)
            if self._is_new():
                dict_.update(self._request.GET)
            if dest in self._request.dbsession:
                self._request.dbsession.expunge(dest)
                self._request.dbsession.expire_all()

        self._populate_widgets(form.schema)
        kwargs = {
            "request": self._request,
            "actions": self._item_actions(dest),
            "msg_col": [self.MSG_COL["copy_ok"]],
        }

        return {
            "title": form.title,
            "form": form,
            "form_render_args": (dict_,),
            "form_render_kwargs": kwargs,
            "deform_dependencies": form.get_widget_resources(),
        }

    def duplicate(self) -> ObjectResponse:
        src = self._get_object()
        return self.copy(src)

    def save(self) -> SaveResponse:
        obj = self._get_object()
        try:
            form = self._form()
            self._populate_widgets(form.schema)
            form_data = self._request.POST.items()
            self._appstruct = form.validate(form_data)
            with self._request.dbsession.no_autoflush:
                obj = form.schema.objectify(self._appstruct, obj)
            self._obj = self._request.dbsession.merge(obj)
            self._request.dbsession.flush()
            return HTTPFound(
                self._request.route_url(
                    "c2cgeoform_item",
                    action="edit",
                    id=self._obj.__getattribute__(self._id_field),  # type: ignore[arg-type] # pylint: disable=unnecessary-dunder-call
                    _query=[("msg_col", "submit_ok")],
                )
            )
        except ValidationFailure as exception:
            self._populate_widgets(form.schema)
            kwargs = {"request": self._request, "actions": self._item_actions(obj), "obj": obj}
            return {
                "title": form.title,
                "form": exception,
                "form_render_args": [],
                "form_render_kwargs": kwargs,
                "deform_dependencies": form.get_widget_resources(),
            }

    def delete(self) -> DeleteResponse:
        obj = self._get_object()
        self._request.dbsession.delete(obj)
        self._request.dbsession.flush()
        return {"success": True, "redirect": self._request.route_url("c2cgeoform_index")}



import pyramid.request
from pyramid.view import notfound_view_config

from c2cgeoform import JSONDict


@notfound_view_config(renderer="../templates/404.jinja2")  # type: ignore[misc]
def notfound_view(request: pyramid.request.Request) -> JSONDict:
    request.response.status = 404
    return {}



import json

from colander import Invalid, null
from geoalchemy2 import WKBElement
from geoalchemy2.shape import from_shape, to_shape

from c2cgeoform.tests import DatabaseTestCase


class TestGeometry(DatabaseTestCase):
    def test_serialize_null(self):
        from c2cgeoform.ext.colander_ext import Geometry

        geom_schema = Geometry()

        self.assertEqual(null, geom_schema.serialize({}, null))

    def test_serialize_wkb(self):
        from c2cgeoform.ext.colander_ext import Geometry

        geom_schema = Geometry()

        from shapely.geometry.point import Point

        wkb = from_shape(Point(1.0, 2.0))
        self.assertEqual(
            {"type": "Point", "coordinates": [1.0, 2.0]}, json.loads(geom_schema.serialize({}, wkb))
        )

    def test_serialize_reproject(self):
        from c2cgeoform.ext.colander_ext import Geometry

        geom_schema = Geometry(srid=4326, map_srid=3857)

        from shapely.geometry.point import Point

        wkb = from_shape(Point(1.0, 2.0), 4326)
        geo_json = json.loads(geom_schema.serialize({}, wkb))
        self.assertEqual("Point", geo_json["type"])
        self.assertAlmostEqual(111319.49079327231, geo_json["coordinates"][0])
        self.assertAlmostEqual(222684.20850554455, geo_json["coordinates"][1])

    def test_serialize_invalid(self):
        from c2cgeoform.ext.colander_ext import Geometry

        geom_schema = Geometry()

        self.assertRaises(Invalid, geom_schema.serialize, {}, "Point(1 0)")

    def test_deserialize_null(self):
        from c2cgeoform.ext.colander_ext import Geometry

        geom_schema = Geometry()

        self.assertEqual(null, geom_schema.deserialize({}, null))
        self.assertEqual(null, geom_schema.deserialize({}, ""))

    def test_deserialize_valid_geojson(self):
        from c2cgeoform.ext.colander_ext import Geometry

        geom_schema = Geometry()

        from shapely.geometry.point import Point

        expected_wkb = WKBElement(Point(1.0, 2.0).wkb)

        wkb = geom_schema.deserialize({}, '{"type": "Point", "coordinates": [1.0, 2.0]}')
        self.assertEqual(expected_wkb.desc, wkb.desc)

    def test_deserialize_reproject(self):
        from c2cgeoform.ext.colander_ext import Geometry

        geom_schema = Geometry(srid=4326, map_srid=3857)

        wkb = geom_schema.deserialize(
            {}, '{"type": "Point", ' '"coordinates": [111319.49079327231, 222684.20850554455]}'
        )
        self.assertEqual(4326, wkb.srid)

        shape = to_shape(wkb)
        self.assertAlmostEqual(1.0, shape.x)
        self.assertAlmostEqual(2.0, shape.y)

    def test_serialize_invalid_wrong_type(self):
        from c2cgeoform.ext.colander_ext import Geometry

        geom_schema = Geometry()

        self.assertRaises(
            Invalid, geom_schema.deserialize, {}, '{"type": "InvalidType", "coordinates": [1.0, 2.0]}'
        )

    def test_serialize_invalid_syntax(self):
        from c2cgeoform.ext.colander_ext import Geometry

        geom_schema = Geometry()

        self.assertRaises(Invalid, geom_schema.deserialize, {}, '"type": "Point", "coordinates": [1.0, 2.0]}')


class TestBinaryData(DatabaseTestCase):
    def test_serialize_anything(self):
        from c2cgeoform.ext.colander_ext import BinaryData

        binary = BinaryData()
        serialized = binary.serialize({}, b"a string of binary data")
        self.assertNotEqual(null, serialized)
        self.assertEqual(b"a string of binary data", serialized.getvalue())

    def test_serialize_null(self):
        from c2cgeoform.ext.colander_ext import BinaryData

        binary = BinaryData()
        self.assertEqual(null, binary.serialize({}, null))

    def test_serialize_empty_string(self):
        from c2cgeoform.ext.colander_ext import BinaryData

        binary = BinaryData()
        self.assertEqual(null, binary.serialize({}, ""))

    def test_deserialize_null(self):
        from c2cgeoform.ext.colander_ext import BinaryData

        binary = BinaryData()
        self.assertEqual(null, binary.deserialize({}, null))

    def test_deserialize_empty_string(self):
        from c2cgeoform.ext.colander_ext import BinaryData

        binary = BinaryData()
        self.assertEqual(null, binary.deserialize({}, ""))

    def test_deserialize_file(self):
        import os

        from c2cgeoform.ext.colander_ext import BinaryData

        dirpath = os.path.dirname(os.path.realpath(__file__))
        file_ = open(os.path.join(dirpath, "data", "1x1.png"), "br")
        binary = BinaryData()
        self.assertIsInstance(binary.deserialize({}, file_), bytes)

        # test that the file can be read multiple times (simulates that
        # a file in the tmpstore is requested several times)
        self.assertEqual(95, len(binary.deserialize({}, file_)))
        self.assertEqual(95, len(binary.deserialize({}, file_)))
        self.assertEqual(95, len(binary.deserialize({}, file_)))



import warnings
from typing import Any, Optional

import pyramid.config
import pyramid.request
from pyramid.config.actions import action_method

from c2cgeoform import JSONDict


class Application:
    """
    Class used to register and access applications details.

    Example usage

    .. code-block:: python

        application_name = request.application.name()
    """

    def __init__(
        self, name: str, models: list[tuple[str, type[Any]]], url_segment: Optional[str] = None
    ) -> None:
        self._name = name
        self._models = models
        self._url_segment = url_segment or name

    def name(self) -> str:
        return self._name

    def url_segment(self) -> str:
        return self._url_segment

    def tables(self) -> list[JSONDict]:
        tables: list[JSONDict] = []
        for key, mapper in self._models:
            ca_config = getattr(mapper, "__colanderalchemy_config__", {})
            table: JSONDict = {}
            # key stand for url path, sometimes designated as table (in match dict)
            table["key"] = key
            table["title"] = ca_config.get("title", mapper.__tablename__)
            table["plural"] = ca_config.get("plural", mapper.__tablename__)
            tables.append(table)
        return tables


class ApplicationRoutePredicate:
    """
    Internal route predicate which checks application segment match a
    registered application.
    """

    def __init__(self, val: Any, config: pyramid.config.Configurator):
        del config  # unused
        self._val = val

    def text(self) -> str:
        return f"c2cgeoform_application = {self._val}"

    phash = text

    def __call__(self, context: dict[str, Any], request: pyramid.request.Request) -> bool:
        app_segment = context["match"].get("application", "default")
        for application in request.registry["c2cgeoform_applications"]:
            if application.url_segment() == app_segment:
                request._c2cgeoform_application = application
                return True
        return False


def pregenerator(
    request: pyramid.request.Request, elements: list[str], kwargs: dict[str, Any]
) -> tuple[list[str], dict[str, Any]]:
    """
    Route pregenerator that set the current matched route application and table
    segments as defaults for generating urls.
    """
    if "application" not in kwargs:
        kwargs["application"] = request.matchdict.get("application", "default")
    if "table" not in kwargs:
        kwargs["table"] = request.matchdict.get("table", None)
    return elements, kwargs


@action_method  # type: ignore[misc]
def add_c2cgeoform_application(
    config: pyramid.config.Configurator,
    name: str,
    models: list[tuple[str, type[Any]]],
    url_segment: Optional[str] = None,
) -> None:
    def register_application() -> None:
        config.registry["c2cgeoform_applications"].append(Application(name, models, url_segment=url_segment))

    config.action(("c2cgeoform_application", name), register_application)


def get_application(request: pyramid.request.Request) -> Optional[Application]:
    return getattr(request, "_c2cgeoform_application", None)


def includeme(config: pyramid.config.Configurator) -> None:
    config.registry.setdefault("c2cgeoform_applications", [])
    config.add_directive("add_c2cgeoform_application", add_c2cgeoform_application)
    config.add_route_predicate("c2cgeoform_application", ApplicationRoutePredicate)
    config.add_request_method(get_application, "c2cgeoform_application", reify=True)


def register_route(config: pyramid.config.Configurator, route: str, pattern: str) -> None:
    config.add_route(
        route,
        pattern,
        pregenerator=pregenerator,
        c2cgeoform_application="registered_application",
    )


def register_routes(config: pyramid.config.Configurator, multi_application: bool = True) -> None:
    if multi_application:
        base_route = "/{application}/{table}"
        register_route(config, "c2cgeoform_locale", "/{application}/locale")
    else:
        base_route = "/{table}"
        register_route(config, "c2cgeoform_locale", "/locale")

    register_route(config, "c2cgeoform_index", base_route)
    register_route(config, "c2cgeoform_grid", f"{base_route}/grid.json")
    register_route(config, "c2cgeoform_map", f"{base_route}/map")
    register_route(config, "c2cgeoform_geojson", f"{base_route}/geojson.json")
    register_route(config, "c2cgeoform_item", f"{base_route}/{{id}}")
    register_route(config, "c2cgeoform_item_duplicate", f"{base_route}/{{id}}/duplicate")


def register_models(
    config: pyramid.config.Configurator, models: type[Any], url_segment: Optional[str] = None
) -> None:
    """
    Deprecated, use config.add_c2cgeoform_application instead.
    """
    warnings.warn(
        'The "register_models" method is deprecated. Use "config.add_c2cgeoform_application" instead.',
        DeprecationWarning,
        stacklevel=3,
    )
    config.add_c2cgeoform_application("default", models, url_segment=url_segment)
    register_routes(config, False)



# pylint: disable=no-self-use

from unittest import TestCase

import pytest
from pyramid import testing
from pyramid.config import Configurator
from pyramid.exceptions import ConfigurationConflictError
from pyramid.interfaces import IRoutesMapper

from c2cgeoform.routes import (
    Application,
    ApplicationRoutePredicate,
    get_application,
    register_models,
    register_routes,
)


class TestDirectives(TestCase):
    def test_add_c2cgeoform_application(self):
        config = Configurator()
        config.include("c2cgeoform.routes")
        config.add_c2cgeoform_application("app", models=[])
        self.assertEqual(0, len(config.registry["c2cgeoform_applications"]))
        config.commit()
        registered_applications = config.registry["c2cgeoform_applications"]
        self.assertEqual(1, len(registered_applications))
        self.assertEqual("app", registered_applications[0].name())

    def test_add_c2cgeoform_application_conflict(self):
        config = Configurator()
        config.include("c2cgeoform.routes")
        config.add_c2cgeoform_application("app", models=[])
        config.add_c2cgeoform_application("app", models=[])
        with self.assertRaises(ConfigurationConflictError):
            config.commit()


class TestApplicationRoutePredicate(TestCase):
    def _config(self, config, app_name, url_segment=None):
        config.include("c2cgeoform.routes")
        config.add_c2cgeoform_application(app_name, models=[], url_segment=url_segment)
        config.commit()

    def test_not_match(self):
        with testing.testConfig() as config:
            self._config(config, "app")
            pred = ApplicationRoutePredicate(config, "anything")
            context = {"match": {"application": "not_registered_application"}}
            request = testing.DummyRequest()
            self.assertFalse(pred(context, request))

    def test_match(self):
        with testing.testConfig() as config:
            self._config(config, "app")
            pred = ApplicationRoutePredicate(config, "anything")
            context = {"match": {"application": "app"}}
            request = testing.DummyRequest()
            self.assertTrue(pred(context, request))

            # We should have matched application in request.application
            application = get_application(request)
            self.assertTrue(isinstance(application, Application))
            self.assertEqual("app", application.name())

    def test_url_segment_not_match(self):
        with testing.testConfig() as config:
            self._config(config, "app", url_segment="some_segment_url")
            pred = ApplicationRoutePredicate(config, "anything")
            context = {"match": {"application": "app"}}
            request = testing.DummyRequest()
            self.assertFalse(pred(context, request))

    def test_url_segment_match(self):
        with testing.testConfig() as config:
            self._config(config, "app", url_segment="some_segment_url")
            pred = ApplicationRoutePredicate(config, "anything")
            context = {"match": {"application": "some_segment_url"}}
            request = testing.DummyRequest()
            self.assertTrue(pred(context, request))

            # We should have matched application in request.c2cgeoform_application
            application = get_application(request)
            self.assertTrue(isinstance(application, Application))
            self.assertEqual("app", application.name())


class TestPregenerator(TestCase):
    def test_pregenerator(self):
        with testing.testConfig() as config:
            config.include("c2cgeoform.routes")
            register_routes(config)
            config.commit()
            request = testing.DummyRequest()
            request.matchdict = {"application": "app", "table": "table"}
            self.assertEqual("http://example.com/app/table", request.route_url("c2cgeoform_index"))


class TestSingleApp(TestCase):
    @pytest.mark.skip(reason="Not working with Pyramid 2.0")
    def test_register_models(self):
        with testing.testConfig() as config:
            config.include("c2cgeoform.routes")

            class MyModel:
                __tablename__ = "mytable"

            register_models(config, [("mytable", MyModel)])
            register_routes(config, multi_application=False)

            request = testing.DummyRequest(environ={"PATH_INFO": "/mytable"})
            routes_mapper = config.registry.queryUtility(IRoutesMapper)
            info = routes_mapper(request)
            match, route = info["match"], info["route"]
            self.assertEqual("c2cgeoform_index", route.name)
            self.assertEqual("mytable", match["table"])

            # We should have the single application in request.c2cgeoform_application
            application = get_application(request)
            self.assertTrue(isinstance(application, Application))
            self.assertEqual("mytable", application.tables()[0]["key"])

            self.assertEqual("http://example.com/table", request.route_url("c2cgeoform_index", table="table"))



import pyramid.request
import pyramid.response
from pyramid.httpexceptions import HTTPFound
from pyramid.view import view_config


@view_config(route_name="home")  # type: ignore[misc]
def home(request: pyramid.request.Request) -> pyramid.response.Response:
    return HTTPFound(request.route_url("c2cgeoform_index", table="excavations"))



from typing import Any, Tuple, Union

import pyramid.config
from deform import Form, widget
from pkg_resources import resource_filename
from pyramid.config import Configurator
from pyramid.i18n import get_localizer
from pyramid.threadlocal import get_current_request
from translationstring import TranslationStringFactory

_ = TranslationStringFactory("c2cgeoform")

JSON = Union[int, float, str, bool, None, "JSONDict", "JSONList"]  # pylint: disable=invalid-name
JSONDict = dict[str, JSON]
JSONList = list[JSON]


""" Default search paths for the form templates. """
default_search_paths: Tuple[str, ...] = (
    resource_filename("c2cgeoform", "templates/widgets"),
    resource_filename("deform", "templates"),
)


default_map_settings = {
    "srid": 3857,
    "baseLayers": [{"type_": "OSM"}],
    "view": {
        "projection": "EPSG:3857",
        "center": [829170, 5933942],
        "zoom": 7,
    },
    "fitSource": False,
    "fitMaxZoom": 14,
    "focusOnly": False,
    "geolocationTooltip": _("Zoom to current location"),
    "zoomInTipLabel": _("Zoom in"),
    "zoomOutTipLabel": _("Zoom out"),
    "mobile": None,
}


def main(global_config: Any, **settings: Any) -> Any:
    """
    This function returns a Pyramid WSGI application.
    In our case, this is only used for tests.
    """
    del global_config  # unused

    config = Configurator(settings=settings)
    return config.make_wsgi_app()


def includeme(config: pyramid.config.Configurator) -> None:
    """
    Function called when "c2cgeoform" is included in a project (with
    ``config.include('c2cgeoform')``).

    This function creates routes and views for c2cgeoform pages.
    """
    config.include("pyramid_chameleon")
    config.include("pyramid_beaker")  # use Beaker for session storage
    config.include(".routes")
    config.include(".views")
    config.add_static_view("c2cgeoform_static", "static", cache_max_age=3600)
    config.add_static_view("deform_static", "deform:static")
    config.add_translation_dirs("colander:locale", "deform:locale", "locale")

    init_deform(config.root_package.__name__)

    config.scan("c2cgeoform.views")


def translator(term: str) -> str:
    request = get_current_request()
    if request is None:
        return term
    return get_localizer(request).translate(term)  # type: ignore[no-any-return]


def init_deform(root_package: str) -> None:
    Form.set_zpt_renderer(default_search_paths, translator=translator)

    node_modules_root = f"{root_package}:node_modules"

    registry = widget.default_resource_registry
    registry.set_js_resources(
        "typeahead", "0.10.5", f"{node_modules_root}/typeahead.js/dist/typeahead.bundle.min.js"
    )
    registry.set_css_resources("typeahead", "0.10.5", "c2cgeoform:static/js/typeaheadjs.css")

    widget.MappingWidget.fields_template = "mapping_fields"
    widget.FormWidget.fields_template = "mapping_fields"
